import os
from enum import Enum
from typing import Union, List

import pandas
import matplotlib.pyplot as plt
import seaborn as sn

from dsframework.base.trainer.pl_wrapper import *
from trainer.pl_wrapper.network_module import generatedProjectNameNetworkModule

##
# @file
# @brief generatedClass class, based on pytorch lightning 'LightningModule' base class.


class ActionType(Enum):
    TRAINING = 'train'
    VALIDATION = 'val'
    TESTING = 'test'


class StepEpoch(Enum):
    STEP = 'step',
    EPOCH = 'epoch'


class generatedClass(ZIDS_PLModel):
    """! Model template class inherits from ZIDSModel(LightningModule), which organizes your code into 6 sections:

        - Computations (init).
        - Train Loop (training_step)
        - Validation Loop (validation_step)
        - Test Loop(test_step)
        - Prediction Loop (predict_step)
        - Optimizers and LR Schedulers(configure_optimizers)

        Log parameters:
        - For single value, use:
            self.log("test_loss", test_loss)  # prog_bar=True

        - For multiple values use dictionary, use:
            values = {"loss": loss, "acc": acc, "metric_n": metric_n}  # add more items if needed
            self.log_dict(values)

        View results in tensorboard:
            tensorboard --logdir=lightning_logs/

        Loading datasets:

            example:
            @code{.py}
            train_set = datasets.MNIST(os.getcwd() + '/data', download=True, train=True, transform=transform)
            test_set = datasets.MNIST(os.getcwd() + '/data', download=True, train=False, transform=transform)

            # use 20% of training data for validation
            train_set_size = int(len(train_set) * 0.8)
            valid_set_size = len(train_set) - train_set_size

            # split the train set into two
            seed = torch.Generator().manual_seed(42)
            train_set, valid_set = random_split(train_set, [train_set_size, valid_set_size], generator=seed)

            train_set_loader = DataLoader(train_set, num_workers=2)
            val_set_loader = DataLoader(valid_set, num_workers=2)
            test_set_loader = DataLoader(test_set, num_workers=2)
            @endcode


        Training:
            example:
            @code{.py}
            trainer = pl.Trainer(max_epochs=1, callbacks=[checkpoint_callback])
            trainer.fit(model=autoencoder, train_dataloaders=train_set_loader, val_dataloaders=val_set_loader)
            @endcode

        Testing:
            example:
            @code{.py}
            trainer.test(model=autoencoder, dataloaders=test_set_loader)
            @endcode

        Loading a trained model, use:
            - load_from_checkpoint

                example:
                @code{.py}
                model = MyTrainer.load_from_checkpoint('lightning_logs/epoch=0-step=48000.ckpt')
                @endcode


        Saving a model, use:
            - save_checkpoint

                for example:
                @code{.py}
                trainer.save_checkpoint("my_checkpoint.ckpt")
                @endcode

            - ModelCheckpoint - define automated checkpoint saving, use:
                for example:
                @code{.py}
                checkpoint_callback = ModelCheckpoint(dirpath="lightning_logs/", save_top_k=2, monitor="val_loss")
                @endcode


    """
    def __init__(self, model_config=None, metrics_config=None, trainer_config=None, save_path=''):
        """! Model class initializer, receives the network and loss function, it initializes all parameter to be
        tracked. """
        super().__init__(model_config, metrics_config, trainer_config)
        self.test_conf_mat = None

        self.model, self.tokenizer = generatedProjectNameNetworkModule.load_define_model()
        self.batch_size = self.model_config.get('batch_size', 32)
        self.num_classes = self.model_config.get('num_classes', None)
        self.learn_rate = self.model_config.get('learning_rate', 1e-3)

        self._device = self.model_config.get('device', 'cpu')
        self.test_results = None
        self.save_path = save_path

        if self.metrics:
            self.initialize_metrics()

        # self.save_hyperparameters()

    def configure_optimizers(self):
        """! Choose what optimizers and learning-rate schedulers to use in your optimization. Normally you’d need one.
                But in the case of GANs or similar you might have multiple.

                    Returns: Any of this 6 options
                        - Single optimizer.
                        - List or Tuple of optimizers.
                        - Two lists - The first list has multiple optimizers, and the second has multiple LR schedulers
                            (or multiple lr_scheduler_config).
                        - Dictionary, with an "optimizer" key, and (optionally) a "lr_scheduler" key whose value
                            is a single LR scheduler or lr_scheduler_config.
                        - Tuple of dictionaries as described above, with an optional "frequency" key.
                        - None - Fit will run without any optimizer.

                Example:
                    @code{.py}
                    optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
                    return optimizer
                    @endcode

                Another example:
                    @code{.py}
                    SGD_optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.7)
                    return SGD_optimizer
                    @endcode

        """
        optimizer = Optimizers.Adam(self.parameters(), lr=self.learn_rate)

        scheduler = transformers.get_scheduler(
            "linear",
            optimizer=optimizer,
            num_warmup_steps=self.model_config['num_warmup_steps'],
            num_training_steps=self.trainer.estimated_stepping_batches
        )

        return [optimizer], [scheduler]

    def forward(self, x):
        """! Defines the computation performed at every call.

        Implementation example:
            @code{.py}
            x = x.view(x.size(0), -1)
            x = self.nnetwork(x)
            return x
            @endcode
        """
        pass

    def training_step(self, batch, batch_idx):
        """! Override to enable training loop

        Implementation example:
            @code{.py}
            def training_step(self, batch, batch_idx):
                x, y = batch
                x = x.view(x.size(0), -1)
                z = self.encoder(x)
                x_hat = self.decoder(z)
                loss = function.mse_loss(x_hat, x)
                return loss
            @endcode

        If you need to do something with all the outputs of each training_step(), override the
        training_epoch_end() method.
        """
        loss, labels, predictions = self.do_step(batch)

        self.log_metrics(predictions, labels, ActionType.TRAINING, StepEpoch.STEP)
        self.log("train_loss", loss, prog_bar=True, logger=True, on_step=False, on_epoch=True,
                 batch_size=self.batch_size)

        return loss

    def validation_step(self, batch, batch_idx):
        """! Override to enable validation loop

        Implementation example:
            @code{.py}
            def validation_step(self, batch, batch_idx):
                x, y = batch
                x = x.view(x.size(0), -1)
                z = self.encoder(x)
                x_hat = self.decoder(z)
                test_loss = function.mse_loss(x_hat, x)
                self.log("val_loss", test_loss, prog_bar=True)
            @endcode
        """
        loss, labels, predictions = self.do_step(batch)

        self.log_metrics(predictions, labels, ActionType.VALIDATION, StepEpoch.STEP)
        self.log("val_loss", loss, batch_size=self.batch_size)

        return {'loss': loss, 'labels': labels, 'predictions': predictions}

    def validation_epoch_end(self, outputs: Union[EPOCH_OUTPUT, List[EPOCH_OUTPUT]]) -> None:

        labels, preds = self.prepare_results(outputs)

        confusion_matrix = self.val_confusion_matrix(preds, labels)

        self._save_log_conf_mat_image(confusion_matrix, self.current_epoch, self.save_path,
                                      self.model_config['label_names'])

        self._save_report(confusion_matrix, self.current_epoch, self.save_path, self.model_config['label_names'])

    def test_step(self, batch, batch_idx):
        """! Override to enable test loop.

        Implementation example:
            @code{.py}
            x, y = batch
            x = x.view(x.size(0), -1)
            z = self.encoder(x)
            x_hat = self.decoder(z)
            test_loss = function.mse_loss(x_hat, x)
            self.log("test_loss", test_loss)  # prog_bar=True

            return {'loss': loss, 'preds': pred, 'target': y}  # will be available in test_epoch_end
            @endcode
        """
        loss, labels, predictions = self.do_step(batch)

        self.log_metrics(predictions, labels, ActionType.TESTING, StepEpoch.STEP)

        self.log("test_loss", loss, batch_size=self.batch_size)

        return {'loss': loss, 'labels': labels, 'predictions': predictions}

    def test_epoch_end(self, outputs):
        """! Implement this to calculate test results and fill the local dictionary return self.test_results

            Args:
                outputs: Array (ie. dict, list etc.) with accumulated results returned from:
                 def test_step(self, batch, batch_idx)

        Confusion matrix example:
            preds = torch.cat([tmp['preds'] for tmp in outputs])
            targets = torch.cat([tmp['target'] for tmp in outputs])

            conf_matrix_results = self.confusion_matrix(preds, targets)

        Results structure example:
            self.test_results = {
                'precision': precision_result,
                'accuracy': accuracy_result,
                'recall': recall_result,
                'f1': f1_result,
                'confusion_matrix': conf_matrix_results}

        """
        labels, preds = self.prepare_results(outputs)

        self.test_conf_mat = self.test_confusion_matrix(preds, labels)

        self._save_log_conf_mat_image(self.test_conf_mat, -1, self.save_path, self.model_config['label_names'])
        self._save_report(self.test_conf_mat, -1, self.save_path, self.model_config['label_names'])

        self.test_results = self.prepare_test_results()

    def log_metrics(self, pred=None, y=None, step_name: ActionType = ActionType.TRAINING,
                    step_epoch: StepEpoch = StepEpoch.STEP):
        """Log the metrics specified in config.py metrics section."""

        for key in self.metrics:

            if key == 'confusion_matrix':
                continue

            if not self.metrics[key]:
                continue

            prog_bar = True if key != 'confusion_matrix' else False

            method_name = "%s_%s" % (ActionType[step_name.name].value, key)

            try:
                metric_method = getattr(self, method_name)
            except AttributeError as e:
                print(f'No method {method_name} metric found, exception: {e}.')
                return

            if step_epoch == StepEpoch.STEP:
                metric_method(pred, y)
                self.log(f'{method_name}', metric_method, prog_bar=prog_bar, on_step=False, on_epoch=True,
                         batch_size=self.batch_size)
            else:
                self.log(f'{method_name}_epoch', metric_method, prog_bar=prog_bar, on_step=False, on_epoch=True,
                         batch_size=self.batch_size)

    def prepare_test_results(self):
        """Prepare test results' dictionary with configured metrics."""
        test_results = {}

        for key in self.metrics:

            if key == 'confusion_matrix':
                test_results[key] = self.test_conf_mat
                continue

            if not self.metrics[key]:
                continue

            method_name = "%s_%s" % ('test', key)
            metric_method = getattr(self, method_name)
            test_results[key] = metric_method.compute()

        return test_results

    def get_test_results(self):
        """Returns test_results."""
        if self.test_results:
            return self.test_results
        else:
            return None

    def do_step(self, batch):

        input_ids = batch[0].to(self._device)
        # input_mask = batch[1].to(self._device)
        input_labels = batch[1].to(self._device)
        input_ids = input_ids.view(input_ids.size(0), -1)
        logits = self.model(input_ids)

        loss = self.loss_function(logits, input_labels)

        labels = input_labels.view(-1)
        predictions = torch.argmax(logits, dim=1)

        return loss, labels, predictions

    def prepare_results(self, outputs):

        preds = torch.cat([tmp['predictions'] for tmp in outputs])
        labels = torch.cat([tmp['labels'] for tmp in outputs])

        return labels, preds

    def _save_log_conf_mat_image(self, confusion_matrix, current_epoch, save_path, label_names=None):

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        title = f'Confusion matrix, epoch: {current_epoch}' if current_epoch != -1 else f'Confusion matrix - test'
        image_name = f'conf_mat_{current_epoch}.png' if current_epoch != -1 else 'conf_mat_test.png'

        if label_names is None:
            df_conf_mat = pandas.DataFrame(confusion_matrix.numpy())
        else:
            df_conf_mat = pandas.DataFrame(confusion_matrix.numpy(), index=label_names, columns=label_names)

        plt.figure(figsize=(10, 7))
        plt.title(title)
        cm = sn.heatmap(df_conf_mat, annot=True, fmt='d')
        figure = cm.get_figure()
        figure.savefig(os.path.join(save_path, image_name))
        plt.close()

        self.trainer_config['logger'].experiment.add_figure("Confusion matrix", figure, self.current_epoch)

    def _save_report(self, confusion_matrix, current_epoch, save_path, label_names=None):

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        file_name = f'metrics_report_{current_epoch}.text' if current_epoch != -1 else 'metrics_report.text'
        file_name = os.path.join(save_path, file_name)

        report_string = f'\nEpoch: {current_epoch}\n' \
                        f'Confusion Matrix:\n{confusion_matrix}\n'
        if label_names:
            report_string = report_string + f'Number of Tokens:\n{len(label_names)}\n'

        with open(file_name, "a") as f:
            f.write(report_string)

    def initialize_metrics(self):

        average = self.metrics.get('average', 'micro')

        for key in self.metrics:
            if not self.metrics[key]:
                continue

            if key == 'accuracy':
                self.train_accuracy = torchmetrics.Accuracy(average=average, num_classes=self.num_classes)
                self.val_accuracy = torchmetrics.Accuracy(average=average, num_classes=self.num_classes)
                self.test_accuracy = torchmetrics.Accuracy(average=average, num_classes=self.num_classes)
            elif key == 'precision':
                self.train_precision = torchmetrics.Precision(average=average, num_classes=self.num_classes)
                self.val_precision = torchmetrics.Precision(average=average, num_classes=self.num_classes)
                self.test_precision = torchmetrics.Precision(average=average, num_classes=self.num_classes)
            elif key == 'recall':
                self.train_recall = torchmetrics.Recall(average=average, num_classes=self.num_classes)
                self.val_recall = torchmetrics.Recall(average=average, num_classes=self.num_classes)
                self.test_recall = torchmetrics.Recall(average=average, num_classes=self.num_classes)
            elif key == 'f1':
                self.train_f1 = torchmetrics.F1Score(average=average, num_classes=self.num_classes)
                self.val_f1 = torchmetrics.F1Score(average=average, num_classes=self.num_classes)
                self.test_f1 = torchmetrics.F1Score(average=average, num_classes=self.num_classes)
            elif key == 'confusion_matrix':
                if not self.model_config['num_classes_conf_mat']:
                    raise Exception(
                        "To use confusion matrix metric, please set 'num_classes_conf_mat' parameter in config.py.")
                self.train_confusion_matrix = torchmetrics.ConfusionMatrix(
                    num_classes=self.model_config['num_classes_conf_mat'])
                self.val_confusion_matrix = torchmetrics.ConfusionMatrix(
                    num_classes=self.model_config['num_classes_conf_mat'])
                self.test_confusion_matrix = torchmetrics.ConfusionMatrix(
                    num_classes=self.model_config['num_classes_conf_mat'])




class ZIDSTrainerTest:

    trainer = None
    model = None
    data_module = None

    def __init__(self, model_path=None, test_dataset_path=None, trainer_config=None):
        self.model_path = model_path
        self.test_dataset_path = test_dataset_path
        self.trainer_config = trainer_config

        self.load_model(model_path)
        self.create_data_module(self.model.tokenizer, test_dataset_path)
        self.trainer_setup(self.trainer_config)

    def load_model(self, model_path):
        pass

    def create_data_module(self, tokenizer, test_dataset_path):
        pass

    def trainer_setup(self, trainer_config):
        pass

    def execute_test(self):
        pass



import hashlib
import json
import os
import sys
import unittest

from dsframework.base.server.cache.cache import Cache, CacheItem
from dsframework.base.server.cache.cache_utils import CacheProvider, DoNothingCache, InMemoryCache, CacheFacade, \
    Compression, KeyTransformer
from dsframework.base.server.cache.redis_cache import RedisCache


class TestCacheProvider(unittest.TestCase):
    def test_get_cache_none(self):
        cache = CacheProvider.get_cache(Cache.Type.NONE)
        self.assertIsInstance(cache, DoNothingCache)

    def test_get_cache_redis(self):
        cache = CacheProvider.get_cache(Cache.Type.REDIS)
        self.assertIsInstance(cache, RedisCache)

    def test_get_cache_in_memory(self):
        cache = CacheProvider.get_cache(Cache.Type.IN_MEMORY)
        self.assertIsInstance(cache, InMemoryCache)

    def test_get_cache_invalid_type(self):
        with self.assertRaises(AttributeError):
            CacheProvider.get_cache("InvalidType")


class TestCacheFacade(unittest.TestCase):
    def setUp(self):
        self.mock_cache = InMemoryCache()

    def test_set_get(self):
        # Test with full cache hit
        input_ = [{"key": "value1"}, {"key": "value2"}]
        self.cache_facade = CacheFacade(self.mock_cache, Cache.Type.IN_MEMORY,
                                        input_, "test_host", "test_service_account")
        self.cache_facade.get("key")
        self.cache_facade.set(input_)
        result = self.cache_facade.get("key")
        self.assertEqual(result, input_)

        # Test with partial cache hit should return empty
        partial_input = [{"key": "value1"}, {"key": "value_not_in_cache"}]
        self.cache_facade = CacheFacade(self.mock_cache, Cache.Type.IN_MEMORY,
                                        partial_input, "test_host", "test_service_account")
        result = self.cache_facade.get("key")
        self.assertEqual([], result)


class TestCompression(unittest.TestCase):
    def test_pack_predictable(self):
        value = [1, 2, 3]
        packed_value = Compression.pack_predictable(value)

        # Ensure that the packed value is not the same as the original value
        self.assertNotEqual(packed_value, value)

        # Ensure that the packed value can be decompressed and loaded correctly
        self.assertEqual(Compression.unpack_predictable(packed_value), value)


class TestKeyTransformer(unittest.TestCase):

    def test_serialize(self):
        key = {'a': 1, 'b': 2}
        key_transformer = KeyTransformer()
        serialized_key = key_transformer._serialize(key)

        # Ensure that the serialized key is a JSON string
        self.assertIsInstance(serialized_key, str)
        try:
            json.loads(serialized_key)
        except ValueError:
            self.fail('_serialize did not produce a proper JSON as expected')

        # Ensure that the serialized key is not the same as the original key
        self.assertNotEqual(serialized_key, key)

    def test_build_key_string(self):
        key = 'key'
        service_name = 'service'
        git_tag_version = 'v1.0'
        os.environ['GAE_SERVICE'] = service_name
        os.environ['DD_VERSION'] = git_tag_version
        key_transformer = KeyTransformer()
        built_key_string = key_transformer._build_key_string(key)

        # Ensure that the built key string is the correct format
        self.assertEqual(built_key_string, f'{service_name}_{git_tag_version}_{key}')

    def test_encode(self):
        key = 'key'
        key_transformer = KeyTransformer()
        encoded_key = key_transformer._encode(key)

        # Ensure that the encoded key is a bytes object
        self.assertEquals(encoded_key, key.encode('utf-8', 'replace'))

    def test_hash(self):
        key = b'key'
        key_transformer = KeyTransformer()
        hashed_key = key_transformer._hash(key)

        # Ensure that the hashed key is a string
        self.assertEquals(hashed_key, hashlib.sha256(key).hexdigest())



from dsframework.base.trainer.pl_wrapper import *


class ZIDSIterableDataset(IterableDataset):
    """! This is the base class for iterable datasets, override __iter__ method."""

    def __init__(self, data):
        self.data_source = data

    def __iter__(self):
        return self.data_source


def worker_init_fn(worker_id):
    pass



# ======= Torch imports =========
import torch
from torch import nn
import torch.nn.functional as F
from torch.utils.data import DataLoader, Dataset, IterableDataset, random_split, get_worker_info
import torchmetrics
import transformers
from torchvision import transforms, datasets
import pytorch_lightning as pl
from pytorch_lightning import Callback
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
from pytorch_lightning.loggers import LightningLoggerBase, TensorBoardLogger
from pytorch_lightning.accelerators import Accelerator
from pytorch_lightning.profiler import BaseProfiler
from pytorch_lightning.strategies import Strategy
from pytorch_lightning.plugins import PLUGIN_INPUT
from pytorch_lightning.utilities.types import (
    STEP_OUTPUT,
    EPOCH_OUTPUT,
    EVAL_DATALOADERS,
    TRAIN_DATALOADERS
)

# ======== DSF imports ========
from dsframework.base.trainer.pl_wrapper.optimizers import Optimizers
from dsframework.base.trainer.pl_wrapper.model.plmodel import ZIDS_PLModel
from dsframework.base.trainer.pl_wrapper.torch_module import ModuleBase
from dsframework.base.trainer.pl_wrapper.data.data_module import ZIDS_PLDataModule
from dsframework.base.trainer.pl_wrapper.data.custom_dataset import ZIDS_PLCustomDataset
from dsframework.base.trainer.pl_wrapper.data.iterable_dataset import ZIDSIterableDataset
from dsframework.base.trainer.pl_wrapper.model.nnetwork import ZIDSNetwork

# tensorboard --logdir=dsframework/cli/trainer/lightning_logs



from dsframework.base.trainer.pl_wrapper import *


class ZIDS_PLCustomDataset(Dataset):
    """! Base class for loading an external custom dataset"""
    x_data = []
    y_data = []

    def __init__(self, x, y):
        self.x_data = x
        self.y_data = y

    def __len__(self):
        return len(self.y_data)

    def __getitem__(self, index):
        return self.x_data[index], self.y_data[index]



import json
from typing import List, Optional
from pydantic import validator
from pydantic.main import BaseModel

##
# @file
# @brief Test the input row data from the DSP.
#        Data row from the DSP -> DSF.

class TestInputRow(BaseModel):
    """! BaseModel (pydantic.main) the base class for TestInputRow.
    @verbatim
    Args:
        data(str) : The data of a specific row, formatted into a json string
        id(int) : Dataset ID from the portal
        raw_id: Optional[int] = -1
    @endverbatim
    """
    data: str
    id: int
    raw_id: Optional[int] = -1

    @validator('data')
    def data_must_be_valid_json(cls, v):
        """! Method that check the data if it's valid json string.

        Args:
            cls : not in use
            v : Data that is going to be checked.
        Raises:
            ValueError
        """
        try:
            json.loads(v)
        except:
            raise ValueError('data must be a valid json string')
        return v


class TestInputRequest(BaseModel):
    """! BaseModel (pydantic.main) the base class for TestInputRequest.
    @verbatim
    Args:
        truth_dataset_id: int = -1 : ID of the truth dataset
        model_type_id(=-1) : Model ID as defined by a table held in staging DSP (number per project)
        rows: List[TestInputRow] : List of TestInputRow
    @endverbatim
    """
    truth_dataset_id: int = -1
    model_type_id: int = -1
    rows: List[TestInputRow]



from enum import Enum
from dsframework.base.trainer.pl_wrapper import *


class Optimizers(Enum):
    """! Wraps the most commonly used optimization algorithm. Direct usages of the optimizer is possible using
    torch.optim. class.
    """
    SGD = (torch.optim.SGD,)
    Adam = (torch.optim.Adam,)
    Adadelta = (torch.optim.Adadelta,)
    Adagrad = (torch.optim.Adagrad,)
    AdamW = (torch.optim.AdamW,)
    SparseAdam = (torch.optim.SparseAdam,)  # Implements lazy version of Adam algorithm suitable for sparse tensors.
    Adamax = (torch.optim.Adamax,)  # Implements Adamax algorithm (a variant of Adam based on infinity norm).
    ASGD = (torch.optim.ASGD,)  # Implements Averaged Stochastic Gradient Descent.
    LBFGS = (torch.optim.LBFGS,)
    NAdam = (torch.optim.NAdam,)
    RAdam = (torch.optim.RAdam,)
    RMSprop = (torch.optim.RMSprop,)
    Rprop = (torch.optim.Rprop,)  # Implements the resilient backpropagation algorithm.

    def __call__(self, *args, **kwargs):
        return self.value[0](*args, **kwargs)



import sys
import wget
import subprocess
import os
import pathlib
type = ''
if sys.argv and len(sys.argv) > 1:
    type = sys.argv[1]
if not type:
    raise Exception('type must be enter')
filename = ''
url = ''
if type == 'windows':
    filename = 'doxygen-1.9.4-setup.exe'
    url = f'https://www.doxygen.nl/files/{filename}'
if type == 'linux':
    filename = 'install.sh'
    url = f'https://raw.githubusercontent.com/Homebrew/install/HEAD/{filename}'
if not filename:
    raise Exception('type not supported')
if not os.path.exists(filename):
    filename = wget.download(url)
path = pathlib.Path(__file__).parent.resolve()
finalPath = os.path.join(path, filename)
process = [finalPath]
if type == 'linux':
    process = [f'chmod +x {finalPath}'] + process
p = subprocess.Popen(process, bufsize=2048, shell=True,
        stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True)
p.wait()
result_str = p.stdout.read()
os.remove(filename)



"""! @brief ZIDS_Stage base class for the batch stage class."""
import logging
import time
from typing import Mapping, OrderedDict
from pyspark.sql import SparkSession

##
# @file
# @brief Defines stage base class.
class ZIDS_BatchPipelineStage():
    def __init__(self, stage_config, logger):
        """! ZIDS_BatchPipelineStage initializer.
        Loads config file.

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """
        self.logger = logger
        logger.info(f"Initializing Stage: {self.get_name()} ")

        self.bucket_name = stage_config['bucket_name']
        self.project_id = stage_config['project_id']
        self.project_name = stage_config['project_name']
        self.region = stage_config['region']

        user_email = ""
        if 'user_email' in stage_config:
            user_email = stage_config['user_email'] + '/'

        self.unique_iteration_id = stage_config['unique_iteration_id']
        if self.unique_iteration_id:
            print(f'unique_iteration_id: {self.unique_iteration_id}')
            self.folder_path = self.project_name + (user_email if user_email else '') + \
                '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + '/' + (user_email if user_email else '') + self.wf_name + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
        self.start_date = ""
        self.end_date = ""
        self.extra_params = ""

    def get_name(self):
        """! Get the stage name
            Returns:
                The stage name string
        """
        return self.__class__.__name__

    def update_stage_params(self, start_date, end_date, params):
        """! Update start date
            Args:
                start_date: String, containing the starting date, received from Airflow
                end_date: String, containing the end date, received from Airflow
                params: String, containing extra parameters provided by user
        """
        print(f"{start_date = }, {end_date = }, {params = }")
        self.start_date = start_date
        self.end_date = end_date
        self.extra_params = params

    def load(self):
        # Override to load input dataframes, and run input data tests if needed
        raise NotImplementedError

    def run(self):
        # Override to execute transform
        raise NotImplementedError

    def finish(self):
        # Override to save output dataframes, and run output data tests if needed
        raise NotImplementedError

    def __call__(self):
        """Run load(), run(), and finish() and log the runtime"""
        start = time.time()

        stage_name = self.get_name()
        self.logger.info(f"Starting stage: {stage_name} ")
        self.load()
        self.logger.info(f"{stage_name}: Finished load()")
        self.run()
        self.logger.info(f"{stage_name}: Finished run()")
        self.finish()
        self.logger.info(f"{stage_name}: Finished finish()")
        elapsed = time.time() - start

        self.logger.info(f"Stage '{stage_name}' finished in " + '%0.1f min' % (elapsed / 60.0))

    def load_spark(self) -> SparkSession:  # TODO: Generalize this in order to receive config options.
        """! Basic loading of a spark session.

            Returns:
                A basic spark session
        """
        spark = SparkSession.builder \
            .appName(self.project_name + self.get_name()) \
            .config('spark.ui.showConsoleProgress', True) \
            .config("spark.sql.parquet.compression.codec", "gzip") \
            .config('spark.jars', 'gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar') \
            .getOrCreate()

        print(f"Spark ID: {spark.sparkContext.applicationId}")
        return spark


class ZIDS_BatchPipeline:
    """! ZIDS_BatchPipelineStage initializer.
    Loads config file.

        Args:
            stage_config : Configuration dictionary, loaded from configuration file.

    """
    def __init__(self):
        # Override
        self.stages: Mapping[str, ZIDS_BatchPipelineStage] = {}

    def run_stage(self, stage_name: str):
        if stage_name not in self.stages:
            stage_names = '\n'.join(sorted(self.stages.keys()))
            raise KeyError(f"Stage '{stage_name}' not found among stages:\n {stage_names}")

        stage = self.stages[stage_name]
        stage()

    def run_all_stages(self):

        assert type(self.stages) is OrderedDict, "Pipeline self.stages must be an OrderedDict to use run_all_stages()"

        if 'spark' in self.__dir__():
            try:
                # TODO -- not sure if this is needed, since all sparkConf parameters show up in the Spark Job page
                #   and it may be very long and verbose
                logger.info("Spark config:", self.spark.sparkContext.getConf().getAll())
            except:
                # self.spark wasn't a SparkSession object
                pass

        start = time.time()
        for stage_name, stage in self.stages.items():
            self.run_stage(stage_name)
        elapsed = time.time() - start

        logger.info('Finished all stages in %0.1f min' % (elapsed / 60.))



"""! @brief Get Data stage, for obtaining data from a generic source, such as a database."""
import argparse
import sys
import logging
from json import loads as json_loads
from dsframework.base.batch.pipeline_stage_base import ZIDS_BatchPipelineStage

logger = logging.getLogger(__name__)

##
# @file
# @brief Stage main class, implements ZIDS_Stage base class.
class ExampleStage(ZIDS_BatchPipelineStage):
    """! Example Stage class
    Implement a stage that will later be converted to an executable job in a specific workflow.
    """

    def __init__(self, stage_config):
        """! The Stage class (ExampleStage) initializer.
        Base class will load basic configuration parameters, additional fields should be added here

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """

        ##
        # @hidecallgraph @hidecallergraph
        logger.info(f"Initializing Stage: {self.get_name()}")

        self.wf_name = "primary"
        self.highest_price_unit_df = None
        self.df = None
        super().__init__(stage_config, logger)
        self.spark = self.load_spark()
        logger.info(f"Stage {self.get_name()} Initialized")

    def get_name(self):
        """! Get the stage name
        """
        return self.__class__.__name__

    def load(self):
        """! The \'load\' phase is for loading input dataframes and running input data tests if needed
        """
        self.df = self.spark.read. \
            options(header='true', inferSchema='true').csv(f"gs://{self.bucket_name}/example_data/retail_day.csv")

    def run(self):
        """! The \'run\' phase is for the transform logic
        """
        self.df.printSchema()
        self.df.createOrReplaceTempView("sales")
        self.highest_price_unit_df = self.spark.sql("select * from sales where UnitPrice >= 3.0")

    def finish(self):
        """! The \'finish\' phase is for saving output dataframes and running output data tests if needed.
        This is also the place to close a spark session, if such was used.
        """
        self.highest_price_unit_df.write.mode("overwrite").parquet(f"{self.bucket_path}"
                                                                   f"/example_data/highest_prices_{self.project_id}.parquet")
        self.spark.stop()


if __name__ == "__main__":
    """! Executes the stage by instantiating it and calling the __call__ function of the base class.

        Args:
            System argument 1 - Configuration file
            System argument 2 - Start date, received from Airflow
            System argument 3 - End date, received from Airflow
            System argument 4 - Placeholder for future parameters
    """

    # Handle input arguments
    parser = argparse.ArgumentParser(description='Configure and run ExampleStage class')
    parser.add_argument('config', type=str,
                        help='Configuration file')
    parser.add_argument('start_date', type=str, default='',
                        help='Start date, received from Airflow')
    parser.add_argument('end_date', type=str, default='',
                        help='End date, received from Airflow')
    parser.add_argument('params', nargs=argparse.REMAINDER,
                        help='Placeholder for future parameters')
    args = parser.parse_args()

    # Configure stage class and run
    stage = ExampleStage(json_loads(args.config))
    try:
        stage.update_stage_params(args.start_date, args.end_date, args.params)
        stage()
    except Exception as e:
        raise Exception(f" Stage failed with error: {e}")



import os
from google.protobuf.duration_pb2 import Duration
from google.cloud.dataproc_v1 import ClusterControllerClient, Cluster, ClusterConfig
from google.cloud.dataproc_v1 import JobControllerClient, Job, PySparkJob
from google.cloud import storage
from ujson import dumps as ujson_dumps
import subprocess
import sys


##
# @file
# @brief Helper functions for batch projects.


class ZIDSBatchHelper:
    """! General class for a variety of helper functions for batch processing projects
    The capabilities of this class include:
    *) Parsing batch configuration file
    *) Creating a Dataproc Cluster
    *) Uploading jars, stages and wheel archive to a given bucket
    """

    def __init__(self, stage_config):
        """! Initialize the class, primarily fill out configuration parameters

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """
        # Save the entire config file for cluster creation
        self.config = stage_config

        self.user_email = ''
        env_type = os.environ.get('SPRING_PROFILES_ACTIVE')
        if (not env_type == 'production') and (not env_type == 'staging'):
            command = 'gcloud config get-value account'
            user_email = subprocess.check_output(command, shell=True).decode(sys.stdout.encoding)
            self.user_email = '/' + user_email.split("@")[0]

        self.wf_name = stage_config['workflow_name']
        self.bucket_name = stage_config['bucket_name']
        self.project_id = stage_config['project_id']
        self.project_name = stage_config['project_name']
        self.region = stage_config['region']
        self.zone = self.region + '-b'  # TODO: Check if we can configure this as well
        self.template_id = stage_config['template_id']
        self.unique_iteration_id = stage_config['unique_iteration_id']
        self.cluster_name = f"{stage_config['cluster_conf']['managed_cluster']['cluster_name']}-{self.wf_name}"
        self.cluster_duration = stage_config['cluster_duration_sec']
        self.managed_cluster = stage_config['cluster_conf']['managed_cluster']

        if self.unique_iteration_id:
            self.folder_path = self.project_name + self.user_email + \
                               '/' + self.wf_name + '/unique_iteration_id_' + self.unique_iteration_id
        else:
            self.folder_path = self.project_name + self.user_email + '/' + self.wf_name + '/main'

        self.bucket_path = f'gs://{self.bucket_name}/{self.folder_path}'
        self.project_path = 'projects/{project_id}/regions/{region}'.format(project_id=self.project_id,
                                                                            region=self.region)
        # Placeholders for file upload actions
        self.storage_client = None
        self.bucket = None

    def create_dataproc_cluster(self):
        """! Create a GCP cluster on DataProc based on configurations from batch_config.json
        """

        # Make sure clust init script will be ready
        self.upload_files_to_bucket('cluster-init')

        # Defined used paths
        main_project_path = os.path.abspath(os.getcwd())

        # Extract pip packages for the cluster
        pip_packages = ''
        with open(main_project_path + '/requirements_docker.txt') as f:
            required = f.read().splitlines()
            required = [x for x in required if "#" not in x]

        with open(main_project_path +
                  f"/batch_files/workflows/{self.wf_name}_workflow/requirements_{self.wf_name}.txt") as f:
            wf_required = f.read().splitlines()
            wf_required = [x for x in wf_required if "#" not in x]

        pip_packages = ' '.join(required + wf_required)

        client = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        # Instantiate a cluster
        cluster = Cluster()
        cluster.cluster_name = self.cluster_name
        cluster_init_action = {"executable_file": f"{self.bucket_path}/batch_files/scripts/cluster-init-actions.sh"}
        cluster_jupyter_support = self.config.get("jupyter_support", False)

        if self.managed_cluster:
            # Configure Jupyter support
            if cluster_jupyter_support:
                from google.cloud.dataproc_v1.types import shared

                software_conf = self.managed_cluster['config']['software_config']

                if 'optional_components' in software_conf and isinstance(software_conf['optional_components'], list):
                    software_conf['optional_components'].append(shared.Component.JUPYTER)
                else:
                    software_conf['optional_components'] = []
                    software_conf['optional_components'].append(shared.Component.JUPYTER)

                self.managed_cluster['config']['endpoint_config'] = {"enable_http_port_access": True}

                self.managed_cluster['config']['config_bucket'] = self.config['cluster_bucket']

            # Configure pip packages
            self.managed_cluster['config']['gce_cluster_config']['metadata']['PIP_PACKAGES'] = pip_packages

            # Configure initialization script
            cluster_init_list = self.managed_cluster['config']['initialization_actions']
            if not isinstance(cluster_init_list, list):
                cluster_init_list = []

            cluster_init_list.append(cluster_init_action)

        managed_cluster_config = self.managed_cluster['config']
        cluster_config = ClusterConfig(managed_cluster_config)

        cluster_duration = Duration(seconds=self.cluster_duration)
        cluster_config.lifecycle_config = {
            'idle_delete_ttl': cluster_duration
        }

        cluster.config = cluster_config

        # Getting cluster if it exists
        exist_cluster = None
        try:
            exist_cluster = client.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=self.cluster_name,
            )
            print(f'Cluster {self.cluster_name} retrieved successfully')

        except Exception as e:
            # A "404", not found exception, is expected here and should only be noted
            if "404" in str(e):
                print(f"Cluster {self.cluster_name} not found. Creating...")
            else:
                print(f'Error getting cluster: {e}')
                raise e

        # Updating the cluster if it exists
        if exist_cluster:
            try:
                updated_config_list = [
                    'config.worker_config.num_instances',
                    'config.secondary_worker_config.num_instances',
                    'labels'
                ]
                client.update_cluster(
                    project_id=self.project_id,
                    region=self.region,
                    cluster_name=self.cluster_name,
                    cluster=cluster,
                    update_mask={
                        "paths": updated_config_list
                    }
                )
                print(f'cluster {self.cluster_name} was updated successfully')

            except Exception as e:
                print(f'error updating cluster: {e}')

        # If no existing cluster was found, create it
        else:
            try:
                client.create_cluster(
                    project_id=self.project_id,
                    region=self.region,
                    cluster=cluster
                )
                print(f'Cluster {self.cluster_name} created successfully')

            except Exception as e:
                print(f'Error creating cluster: {e}')

    def get_client_check_cluster(self) -> (bool, ClusterControllerClient):
        """! Set up a client and check for an existing cluster
            Returns:
                Cluster: If a cluster was found, return its class
                ClusterControllerClient: Client for controlling the cluster
        """

        # Set up a client
        client = ClusterControllerClient(
            client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"}
        )

        # Getting cluster if it exists
        cluster = None
        cluster_exists = False
        try:
            cluster = client.get_cluster(
                project_id=self.project_id,
                region=self.region,
                cluster_name=self.cluster_name,
            )
            print(f'Cluster {self.cluster_name} retrieved successfully')

        except Exception as e:
            # A "404", not found exception, is expected here and should only be noted
            if "404" in str(e):
                print(f"Cluster {self.cluster_name} not found.")
            else:
                print(f'Error getting cluster: {e}')
                raise e
        if cluster:
            cluster_exists = True
        return client, cluster_exists

    def start_dataproc_cluster(self):
        """! Start an existing GCP cluster on DataProc based on configurations from batch_config.json
        """
        client, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                operation = client.start_cluster(
                    request={"project_id": self.project_id,
                             "region": self.region,
                             "cluster_name": self.cluster_name}
                )
                response = operation.result()
                print(f'Started cluster {self.cluster_name}. Result: {response} ')

            except Exception as e:
                print(f'Failed starting cluster {self.cluster_name}')

    def stop_dataproc_cluster(self):
        """! Stop an existing GCP cluster on DataProc based on configurations from batch_config.json
        """

        client, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                operation = client.stop_cluster(
                    request={"project_id": self.project_id,
                             "region": self.region,
                             "cluster_name": self.cluster_name}
                )
                response = operation.result()
                print(f'Stopped cluster {self.cluster_name}. Result: {response} ')

            except Exception as e:
                print(f'Failed stopping cluster {self.cluster_name}')

    def delete_dataproc_cluster(self):
        """! Delete an existing DataProc cluster according to project specific configuration.
        Since we can't update certain parameters on an existing cluster, we need the option
        to delete it
        """

        client, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                operation = client.delete_cluster(
                    project_id=self.project_id,
                    region=self.region,
                    cluster_name=self.cluster_name,
                )
                response = operation.result()
                print(f'Deleted cluster {self.cluster_name}. Result: {response} ')

            except Exception as e:
                print(f'Failed deleting cluster {self.cluster_name}')

    def connect_dataproc_cluster(self):
        """! Connect via ssh to an existing DataProc cluster according to project specific configuration.
        Connection will be made to the master node
        """

        # No need for the "client" variable that returns from the function
        _, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                command = f'gcloud compute ssh {self.cluster_name}-m ' \
                          f'--project={self.project_id} ' \
                          f'--zone={self.zone} ' \
                          f'--tunnel-through-iap'
                process = subprocess.run(command, shell=True, check=True)

                print(f'Connection to cluster {self.cluster_name} ended. Result: {process.returncode}')

            except subprocess.CalledProcessError as e:
                print(f'Failed on connection to cluster {self.cluster_name}, Error: {e}')

    def run_single_stage(self, stage_name: str):
        """! Run a specific batch stage on an existing cluster
            Args:
                stage_name: Stage name to execute as a job
        """

        # Define some job parameters
        stage_dir = f"batch_files/stages/{stage_name}"
        file_name = f"{self.project_name}_{self.wf_name}.zip"

        spark_properties = self.config["spark_properties"]
        if stage_name in spark_properties:
            stage_spark_properties = {**spark_properties['basic'], **spark_properties[stage_name]}
        else:
            stage_spark_properties = spark_properties['basic']

        spark_properties_list = [[key, value] for key, value in list(stage_spark_properties.items())]

        default_jars = ['gs://spark-lib/bigquery/spark-bigquery-latest_2.12.jar']
        default_args = [ujson_dumps(self.config),
                        "start_date_placeholder",
                        "end_date_placeholder",
                        "params_placeholder"]

        _, existing_cluster = self.get_client_check_cluster()

        if existing_cluster:
            try:
                # Create a job client
                client = JobControllerClient(
                    client_options={"api_endpoint": f"{self.region}-dataproc.googleapis.com:443"})

                # Create job to run
                created_job = Job()
                py_job = PySparkJob()

                py_job.main_python_file_uri = f'{self.bucket_path}/{stage_dir}/main.py'
                py_job.archive_uris = [f'{self.bucket_path}/dist/{file_name}']
                py_job.python_file_uris = [f'{self.bucket_path}/dist/{file_name}']
                py_job.properties = spark_properties_list
                py_job.args = default_args
                py_job.jar_file_uris = default_jars

                created_job.pyspark_job = py_job
                created_job.placement.cluster_name = self.cluster_name

                # Submit job
                client.submit_job(
                    project_id=self.project_id,
                    region=self.region,
                    job=created_job
                )
                print(f'Job submitted: {stage_name}.')

            except Exception as e:
                print(f'Failed submitting job {stage_name} to cluster: {self.cluster_name}')
        else:
            print(f"No cluster {self.cluster_name} found. Please create and rerun command")

    def upload_files_to_bucket(self, upload_type: str):
        """! Upload jars, stages, cluster init script and wheel file to a given bucket
                Args:
                    upload_type: String reflecting which file type we wish to upload to
                                 bucket (jars, stages, whl, cluster-init, all)
        """
        valid_types = {'all', 'jars', 'stages', 'whl', 'cluster-init'}
        if upload_type not in valid_types:
            print(f"Can't determine file type {upload_type}. "
                  f"Please use one of: 'all', 'jars', 'stages', 'whl', 'cluster-init'")
            return

        print(f'Uploading files of type: {upload_type} to bucket: {self.bucket_name}')

        self.storage_client = storage.Client(project=self.project_id)
        self.bucket = self.storage_client.get_bucket(self.bucket_name)

        parent_path = os.path.abspath(os.path.join(os.getcwd(), 'batch_files'))
        script_path = os.path.join(parent_path, 'scripts')

        if upload_type == 'all' or upload_type == 'jars':
            self.upload_jars_to_bucket(parent_path)
        if upload_type == 'all' or upload_type == 'stages':
            self.upload_stages_main_to_bucket(parent_path)
        if upload_type == 'all' or upload_type == 'whl':
            self.upload_whl_to_bucket(parent_path)
        if upload_type == 'all' or upload_type == 'cluster-init':
            self.upload_cluster_init_to_bucket(script_path)

    def upload_cluster_init_to_bucket(self, root_dir):
        """! Upload cluster initialization script to a given bucket
                Args:
                    root_dir: Directory in which we search for the cluster initialization script
        """
        name = 'cluster-init-actions.sh'
        path_local = os.path.join(root_dir, name)
        bucket_blob_path = path_local[path_local.index(self.project_name) + len(self.project_name):].replace('\\', '/')
        blob = self.bucket.blob(self.folder_path + bucket_blob_path)
        blob.upload_from_filename(path_local)
        # self.gsutils_upload_one_file(path_local, bucket_blob_path)
        print(f'uploaded cluster init script to: {bucket_blob_path}')

    def upload_jars_to_bucket(self, root_dir):
        """! Upload relevant jars to a given bucket
            Args:
                root_dir: Directory in which we search for the jars directory
        """

        for path, subdirs, files in os.walk(root_dir + '/jars'):
            for name in files:
                if name.endswith('.jar'):
                    path_local = os.path.join(path, name)
                    bucket_blob_path = path_local[path_local.index(self.project_name) +
                                                  len(self.project_name):].replace('\\', '/')
                    blob = self.bucket.blob(self.folder_path + bucket_blob_path)
                    blob.upload_from_filename(path_local)
            # print(path_local)
            # print(bucket_blob_path)
        print('uploaded jars successfully')

    def upload_stages_main_to_bucket(self, root_dir):
        """! Upload relevant stages to a given bucket
            Args:
                root_dir: Directory in which we search for the stages directory
        """

        for path, subdirs, files in os.walk(root_dir + f"/workflows/{self.wf_name}_workflow/stages"):
            for name in files:
                if name == 'main.py':
                    path_local = os.path.join(path, name)
                    stage_name = path_local.rsplit('/', 2)[-2]  # Stage name will always be the parent dir of main.py
                    bucket_blob_path = f"/batch_files/stages/{stage_name}/{name}"
                    blob = self.bucket.blob(self.folder_path + bucket_blob_path)
                    blob.upload_from_filename(path_local)
                # print(path_local)
                # print(bucket_blob_path)
                # print(name)
        print('uploaded stages main successfully')

    def upload_whl_to_bucket(self, root_dir):
        """! Upload relevant wheel file to a given bucket
            Args:
                root_dir: Directory in which we search for the wheel archive
        """
        from concurrent.futures import ThreadPoolExecutor
        import functools

        parent_path = os.path.abspath(os.path.join(root_dir, ".."))
        dist = parent_path + '/dist'
        onlyfiles = [os.path.join(dist, f) for f in os.listdir(dist) if os.path.isfile(os.path.join(dist, f))
                     and self.wf_name in f]

        def copy_file(path_local, project_name, folder_path, dest_bucket):
            bucket_blob_path = path_local[path_local.index(project_name) +
                                          len(project_name):].replace('\\', '/')
            blob = dest_bucket.blob(folder_path + bucket_blob_path)
            blob.chunk_size = 1024 * 1024
            blob.upload_from_filename(path_local)
            return

        # Parallel run. Keep previous method as fallback.
        if True:
            copy_file_part = functools.partial(copy_file, project_name=self.project_name,
                                               folder_path=self.folder_path, dest_bucket=self.bucket)
            with ThreadPoolExecutor(max_workers=5) as executor:
                for filename in onlyfiles:
                    executor.submit(copy_file_part, filename)

        else:
            for path_local in onlyfiles:
                if self.wf_name not in path_local:
                    continue
                bucket_blob_path = path_local[path_local.index(self.project_name) +
                                              len(self.project_name):].replace('\\', '/')
                blob = self.bucket.blob(self.folder_path + bucket_blob_path)
                blob.chunk_size = 1024 * 1024
                blob.upload_from_filename(path_local)
        # print(path_local)
        # print(project_bucket_folder + bucket_blob_path)

        print(f'uploaded package whl successfully to: {self.folder_path}')

    def gsutils_upload_one_file(self, local_path, target_path):
        """! Upload relevant wheel file to a given bucket
            Args:
                local_path: Directory from which we copy the data
                target_path: Directory on the remote location to which we upload the data
        """

        command = '''gsutil -o GSUtil:parallel_composite_upload_threshold=150M -m cp \
        "''' + local_path + '''" \
        "''' + self.bucket_path + target_path + '''" \
        '''

        process = subprocess.call(command, shell=True)




from typing import Optional, Union
from dsframework.base.trainer.pl_wrapper import *
from trainer.pl_wrapper.custom_dataset import generatedProjectNameCustomDataset

##
# @file
# @brief generatedClass class, pytorch lightning 'pl.LightningDataModule' base class.
class generatedClass(ZIDS_PLDataModule):
    """! Template for a Data Module (pytorch lightning 'pl.LightningDataModule' base class.), this is where the
    dataloader is returned for every dataset: training, validation and test."""

    def __init__(self,
                 model_config=None,
                 data_path: str = '',
                 file_type='csv',
                 train_set: Union[Dataset, generatedProjectNameCustomDataset] = None,
                 val_set: Union[Dataset, generatedProjectNameCustomDataset] = None,
                 test_set: Union[Dataset, generatedProjectNameCustomDataset] = None,
                 test_set_out_of_dataset: bool = False,
                 split_percent=0.8):
        super().__init__(train_set=train_set, val_set=val_set, test_set=test_set, model_config=model_config)
        self.train_val_set = None
        self.model_config = model_config
        self.data_path = data_path
        self.file_type = file_type
        self.seed = model_config['seed']
        self.split_percent = split_percent
        self.test_set_out_of_dataset = test_set_out_of_dataset

    def prepare_data(self):
        """! Define steps that should be done on only one GPU, like getting data."""

        if self.data_path != '':
            ds_x, ds_y = self.parse_dataset(self.load_dataset(self.data_path, self.file_type))
            self.train_val_set = generatedProjectNameCustomDataset(ds_x, ds_y)

    def setup(self, stage: Optional[str] = None):
        """! Make assignments here (val/train/test split)
                called on every process in DDP
                Define steps that should be done on
                every GPU, like splitting data, applying
                transform etc.


            Example:
                @code
                # Assign train/val datasets for use in dataloaders:
                if stage in (None, "fit"):
                    mnist_full = MNIST(self.data_dir, train=True, transform=self.transform)
                    self.ds_train, self.ds_val = random_split(mnist_full, [55000, 5000])

                # Assign test dataset for use in dataloader(s):
                if stage in (None, "test"):
                    self.ds_test = MNIST(self.data_dir, train=False, transform=self.transform)

                @endcode
            """
        if self.test_set is None and self.test_set_out_of_dataset:
            self.train_val_set, self.test_set = self.split_dataset(self.train_val_set, self.seed,
                                                                   train_size_percent=self.split_percent)

        if stage in (None, "fit") and self.train_val_set is not None and self.val_set is None:
            self.train_set, self.val_set = self.split_dataset(self.train_val_set, self.seed,
                                                              train_size_percent=self.split_percent)

    def train_dataloader(self):
        """! Return DataLoader for Training Data here

        Example:
            @code
            train_split = Dataset(...)
            return DataLoader(train_split)
            @endcode
        """
        return DataLoader(self.train_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def val_dataloader(self):
        """! Return DataLoader for data validation.

        Example:
            @code
            val_split = Dataset(...)
            return DataLoader(val_split)
            @endcode
        """
        return DataLoader(self.val_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def test_dataloader(self):
        """! Return DataLoader for data testing.

        Example:
            @code
            val_split = Dataset(...)
            return DataLoader(test_split)
            @endcode
        """
        return DataLoader(self.test_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def split_dataset(self, dataset, seed, train_size_percent):
        """! Splits the data based on specified percentage."""

        self.train_val_len = len(dataset)
        len_train = int(self.train_val_len * train_size_percent)
        len_val_test = self.train_val_len - len_train

        return random_split(dataset, [len_train, len_val_test], generator=torch.Generator().manual_seed(seed))

    def parse_dataset(self, remove_header):
        """! Here you need to parse your dataset, return the dataset and target.

        Example:
            @code{.py}
            if remove_header:
                self.dataset = self.dataset[1:]

            x = np.array(self.dataset, dtype=np.float32)[:, :-1]
            y = np.array(self.dataset, dtype=np.float32)[:, -1].reshape(-1, 1)

            return x, y
            @endcode
        """
        raise Exception("Not implemented exception (parse_dataset).")



import logging
import sys
import os
import json
import time

import uvicorn

from typing import List
from pythonjsonlogger import jsonlogger
from fastapi import FastAPI, Depends, Header, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi_utils.tasks import repeat_every
from starlette.requests import Request

from pipeline.pipeline import generatedProjectNamePipeline
from pipeline.schema.inputs import generatedProjectNameInputs
from pipeline.schema.outputs import generatedProjectNameOutputs
from dsframework.base.server.cache.cache import Cache
from dsframework.base.server.cache.cache_utils import CacheProvider, CacheFacade
from tester.test_schema.test_input import TestInputRequest
from tester.test_schema.test_output import TestOutputResponse
from tester.general_tester import generatedProjectNameGeneralTester

##  Logger setup
root = logging.getLogger()
root.setLevel(logging.INFO)
## Redirect logging to sys.stdout
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
## Assign logger a json formatter to output log data as json.
formatter = jsonlogger.JsonFormatter()
handler.setFormatter(formatter)
root.addHandler(handler)

logger = logging.getLogger(__name__)


async def get_token_header(x_token: str = Header(...)):
    if x_token != "fake-super-secret-token":
        raise HTTPException(status_code=400, detail="X-Token header invalid")


async def verify_env():
    # Don't allow usage of an endpoint on production environment
    if os.environ.get('SPRING_PROFILES_ACTIVE') == 'production':
        raise HTTPException(status_code=404, detail="Endpoint not available")


def load_allowed_cors_origins():
    allowed_origins = []
    cors_origin_list_path = 'server/cors_allowed_origins.json'
    if os.path.exists(cors_origin_list_path):
        with open(cors_origin_list_path) as f:
            allowed_origins = json.load(f)

    return allowed_origins


def extract_host_and_service_account(request):
    x_forwarded_for = request.headers.get('x-forwarded-for')
    x_goog_authenticated_user_email = request.headers.get('x-goog-authenticated-user-email')
    service_account = ''
    host = ''
    try:
        if x_goog_authenticated_user_email:
            service_account = x_goog_authenticated_user_email.replace('accounts.google.com:', '')
        if x_forwarded_for:
            host = x_forwarded_for
    except Exception as e:
        print(f'Attempt to get headers from: {request.headers} failed. Error: {e}')
        pass
    return host, service_account


class OutputLogger:
    def __init__(self, endpoint) -> None:
        self.data = None
        self.host = None
        self.service_account = None
        self.endpoint = endpoint

    def save_request_info(self, data, host, service_account) -> None:
        self.data = data
        self.host = host
        self.service_account = service_account

    def log_output(self, output, retrieve_type='', pipeline_timing=None, predictable_object_count=None):

        extra_info = dict(input=self.data, output=output, from_host=self.host,
                          from_service_account=self.service_account)
        if pipeline_timing:
            extra_info.update({"pipeline_exec_time": pipeline_timing,
                               "predictable_object_count": predictable_object_count})
        logger.info(f"INFO {self.endpoint} invoked {retrieve_type}", extra_info)


## Instantiate FastAPI()
app = FastAPI()
## Allowed cors origins
origins = load_allowed_cors_origins()
## Methods allowed, for FastAPI()
methods = ["*"]
## Headers allowed, for FastAPI()
headers = ["*"]
## Credentials required, bool.
credentials = True
##
# @var allow_origins
# Cors allowed origins.
# @var allow_credentials
# Allowed credentials by FastAPI()
# @var allow_methods
# Allowed methods by FastAPI()
# @var allow_headers
# Allowed headers by FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=credentials,
    allow_methods=methods,
    allow_headers=headers,
)

## Initialize caching - turned off by default
## To enable caching change cache_type to Cache.Type.REDIS
cache_type = Cache.Type.NONE
cache = CacheProvider.get_cache(cache_type)
cache_facade = CacheFacade(cache, cache_type)


@app.on_event("startup")
@repeat_every(seconds=60)  # 1 minute
def run_cache_timer() -> None:
    if not cache.status():
        cache.init()


## Initialized a pipeline for /predict /parse endpoints.
pipeline = generatedProjectNamePipeline()

# Initialize logger classes
predict_logger = OutputLogger('predict')


@app.post('/predict', dependencies=[Depends(get_token_header)])
def predict(body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
    """! Predict api endpoint, is designed to support predict projects such as
    scoops classification, valid user email etc.

    Use by post request, for example:

        headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
        url = http://localhost:8080/predict
        data = {}  # The data to run by model.
        response = requests.post(url, json=data, headers=headers)

    """
    host, service_account = extract_host_and_service_account(request)
    data = body.dict()
    cache_facade.get_request_data(data, host, service_account)
    predict_logger.save_request_info(data, host, service_account)

    output = cache_facade.get('--- REPLACE WITH KEYWORD FOR FINDING KEYS INSIDE THE INPUT (data) ---')
    if len(output) > 0:
        predict_logger.log_output(output, 'cache')
        return output
    try:
        tic = time.perf_counter()
        output: List[generatedProjectNameOutputs] = pipeline.execute(**data)  # call model here
        pipeline_timing = f"{time.perf_counter() - tic:0.6f}"
        predict_logger.log_output(output,
                                  pipeline_timing=pipeline_timing,
                                  predictable_object_count=len(output) if isinstance(output, list) else 1)
        cache_facade.set(output)
    except Exception as ex:
        output = {'error': {'request': str(ex)}}
        logger.exception("ERROR predict invoked",
                         extra=dict(input=data, output=output, from_host=host, from_service_account=service_account))
    return output


## Fetch the proper tester for the project
model_tester = generatedProjectNameGeneralTester(pipeline)


@app.post('/test', dependencies=[Depends(get_token_header), Depends(verify_env)], include_in_schema=False)
def test(input_request: TestInputRequest) -> List[TestOutputResponse]:
    """! This is the main endpoint for testing.

        Args:
            input_request: Request as received from the DSP
        Returns:
            A list of responses, to be sent back to the DSP

    Every project starts with a default mock response service, in order to enable DSP integration (when the working
    testing functions are ready and implemented in the tester/general_tester.py),
    replace the mock service with the real one, in practice just remark the following line:

        response = model_tester.create_mock_response(input_request)

    and remove remark quotes from the following block of code:

        model_tester.save_meta_data(input_request)
        response = []
        # Until this model will support batches, work sequentially. Once it does, pass the full list of rows.
        for request_row in input_request.rows:
            input_list = [request_row]
            tester_output = model_tester.test_batch(input_list)
            response.append(tester_output[0])

    """
    response = model_tester.create_mock_response(input_request)

    """    
    model_tester.save_meta_data(input_request)
    response = []
    # Until this model will support batches, work sequentially. Once it does, pass the full list of rows.
    for request_row in input_request.rows:
        input_list = [request_row]
        tester_output = model_tester.test_batch(input_list)
        response.append(tester_output[0])
    """
    return response


@app.get("/livenessprobe")
def liveness_probe():
    """! Endpoint used to test server liveness

        Usage:
            http://localhost:8080/livenessprobe

        Returns:
            server response: {"alive": True} - On success

    """
    return {"alive": True}


@app.get("/readinessprobe")
def readiness_probe():
    """! Endpoint used to test server readiness

        Usage:
            http://localhost:8080/readinessprobe

        Returns:
            server response: {"ready": True} - On success

        """
    return {"ready": True}


if __name__ == '__main__':
    # Runs the server
    ##
    # @var host
    # IP Address
    # @var port
    # Port number
    uvicorn.run(app, host='0.0.0.0', port=8080)



class ZIDSTrainerData:

    dataset_train = None
    dataset_validation = None
    dataset_test = None
    data_module = None
    save_path = ''

    def __init__(self, model_config=None, load_datasets=False, training_path=None, validation_path=None,
                 test_dataset_path=None, tokenizer=None):
        self.model_config = model_config
        self.load_ds = load_datasets
        self.training_path = training_path
        self.validation_path = validation_path
        self.test_dataset_path = test_dataset_path
        self.tokenizer = tokenizer

        if self.load_ds:
            self.load_datasets(self.training_path, self.validation_path, self.test_dataset_path)

        self.create_data_module()

    def create_data_module(self):
        pass

    def get_data_module(self):

        self.data_module = None

        return self.data_module

    def load_datasets(self, ds_training_path=None, ds_val_path=None, ds_test_path=None):
        raise Exception('load_datasets() not implemented.')



import collections
import os
import logging
import pickle
import sys
import zlib
import ujson
import hashlib
from typing import Any, List, Union, Dict
from pydantic import BaseModel
from dsframework.base.server.cache.cache import Cache, CacheItem
from dsframework.base.server.cache.redis_cache import RedisCache

logger = logging.getLogger(__name__)


class KeyTransformer:
    """!Class for transforming cache keys by applying Serialization, Encoding and Hashing by default.
    Inherit and override the transform method to provide a custom transformation implementation.

    Attributes:
        git_tag_version: the git tag of this instance (taken from environment variable 'DD_VERSION')
        service_name: the service name (taken from environment variable 'GAE_SERVICE')
    """

    def __init__(self):
        self.git_tag_version = os.environ.get('DD_VERSION', '')
        self.service_name = os.environ.get('GAE_SERVICE', 'local')

    @staticmethod
    def _serialize(key: dict) -> str:
        """!Returns a JSON serialized key"""
        return ujson.dumps(key, ensure_ascii=False)

    def _build_key_string(self, key: str) -> str:
        """!Returns the key prefixed by the service name and tag version"""
        return f'{self.service_name}_{self.git_tag_version}_{key}'

    @staticmethod
    def _encode(key: str) -> bytes:
        """!Returns a UTF-8 encoded key"""
        return key.encode('utf-8', 'replace')

    @staticmethod
    def _hash(key: bytes):
        """!Returns a SHA256 hashed key"""
        return hashlib.sha256(key).hexdigest()

    def transform(self, key: dict) -> str:
        """!Transforms a given key using a set of pre-defined functions:
        JSON Serialization, UTF-8 Encoding and SHA256 hashing.

        :param key: the key dictionary to transform
        :return: the transformed key string
        """

        result = self._serialize(key)
        result = self._build_key_string(result)
        result = self._encode(result)
        result = self._hash(result)
        return result


class Compression:
    """!Utility class for compressing/decompression values. Applies pickle serialization.
    Inherit this class and override its methods to supply a different compression implementation for the Cache.
    """

    @staticmethod
    def pack_predictable(value):
        """!Pickle dumps + zlib compress (pickle protocol no. 5 for stability)"""
        unzipped = pickle.dumps(value, protocol=5)
        return zlib.compress(unzipped)

    @staticmethod
    def unpack_predictable(value):
        """!Zlib decompress + pickle loads."""
        unzipped = zlib.decompress(value)
        return pickle.loads(unzipped)


default_key_transformer = KeyTransformer()
default_compression = Compression()


class CacheFacade:
    """!Class to encapsulate all the necessities around the basic cache functionalities.
    Saves keys internally on get() to be used later on the set().

    Attributes:
        cache: the actual cache instance
        cache_type: the cache type (enum)
        data: the request data
        host: the host name to be used in logs
        service_account: the service account to be used in logs
        key_transformer: in charge of key transformations (serialization, encoding, hashing by default)
        with_compression: enable/disable value compression (disabled by default)
        compression: compression implementation (pickle + zlib by default)
    """
    def __init__(self, cache: Cache, cache_type: Cache.Type,
                 key_transformer: KeyTransformer = default_key_transformer,
                 with_compression: bool = False, compression: Compression = default_compression) -> None:
        super().__init__()
        self.__cache = cache
        self.__cache_type = cache_type
        self.__data = None
        self.__host = None
        self.__service_account = None
        self.__key_transformer = key_transformer
        self.__compression = compression
        self.__with_compression = with_compression
        self.__cache_keys = None

        # Define the "get" and "set" functions once, so no if statements are needed for every request
        if cache_type is Cache.Type.NONE:
            self.get = self.__cache.get
            self.set = self.__cache.set
        else:
            self.get = self.facade_get
            self.set = self.facade_set

    def get_request_data(self, data, host, service_account) -> None:
        self.__data = data
        self.__host = host
        self.__service_account = service_account

    # noinspection PyBroadException
    def facade_get(self, cache_keys_search_keyword: str) -> List[Any]:
        """!Retrieves one or more items from the cache.
        Transforms each key using the default KeyTransformer or using the supplied one.
        If compression is enabled - decompress values using the default Compression or using the supplied one.

        :param cache_keys_search_keyword: keyword for locating values to be used as cache keys
        :return: a list of one or more items corresponding to the given list of keys
        """
        try:
            input_values = self._get_all_cache_keys_mapped_by_keyword(self.__data, cache_keys_search_keyword)
            # extract the list of signatures to be used as cache keys
            self.__cache_keys = [self.__key_transformer.transform(raw_key) for raw_key in input_values]
            outputs = self.__cache.get(self.__cache_keys)
            if self.all_values_not_none(outputs):
                outputs = [self._decompress_if(output) for output in outputs]
                return outputs
        except Exception:
            logger.exception("ERROR Failed retrieving items from Cache. Calling model instead. ",
                             extra=dict(input=self.__data, from_host=self.__host,
                                        from_service_account=self.__service_account))
        return []

    def _decompress_if(self, item):
        """!Decompress the given item if compression is enabled"""
        return self.__compression.unpack_predictable(item) if self.__with_compression else pickle.loads(item)

    def _get_all_cache_keys_mapped_by_keyword(self, collection: Union[List, Dict], keyword: str, depth: int = 0,
                                              max_depth: int = 10) -> List:
        """!Returns a list of values for the specified key in the collection.

        The function will recursively search through the collection and return a list
        of all values (cache keys) for the specified keyword. If the keyword is not found in the
        collection, the function will return an empty list. The search will stop when
        the maximum depth is reached.

        :param collection: (Union[List, Dict]) The collection to search.
        :param keyword: (str) The keyword to search for.
        :param depth: (int) The current depth of the search.
        :param max_depth: (int) The maximum depth of the search.

        :return: A list of values (cache keys) for the specified keyword.
        """
        values = []
        if depth > max_depth:
            return values
        if isinstance(collection, list):
            for item in collection:
                if isinstance(item, (list, dict)):
                    values.extend(self._get_all_cache_keys_mapped_by_keyword(item, keyword, depth + 1, max_depth))
        elif isinstance(collection, dict):
            if keyword in collection:
                values.append(collection[keyword])
            for value in collection.values():
                if isinstance(value, (list, dict)):
                    values.extend(self._get_all_cache_keys_mapped_by_keyword(value, keyword, depth + 1, max_depth))
        return values

    # noinspection PyBroadException
    def facade_set(self, output: List[BaseModel]) -> None:
        """!Stores one or more items in the cache.
        Transforms each key using the default KeyTransformer or using the supplied one.
        If compression is enabled - values are compressed using the default Compression or using the supplied one.

        :param output: the list of pipeline outputs to be stored
        """

        if len(self.__cache_keys) != len(output):
            # if the number of pipeline outputs is different from the number of keys - don't store the outputs at all
            return

        try:
            cache_items = []
            for i, cache_key in enumerate(self.__cache_keys):
                value = self._compress_if(output[i])
                cache_items.append(CacheItem(cache_key, value))
            self.__cache.set(cache_items)

        except Exception:
            logger.exception("ERROR Failed storing parsed signatures in Redis",
                             extra=dict(input=self.__data, output=output, from_host=self.__host,
                                        from_service_account=self.__service_account))

    def _compress_if(self, item):
        """!Compress the given item if compression is enabled"""
        return self.__compression.pack_predictable(item) if self.__with_compression else pickle.dumps(item)

    @staticmethod
    def all_values_not_none(cache_values: list) -> bool:
        """!Returns True if all values in the given list are not None"""
        return cache_values and all(cache_value is not None for cache_value in cache_values)


class CacheProvider:
    @staticmethod
    def get_cache(cache_type: Cache.Type) -> Cache:
        """!Creates and returns a new cache instance according to the requested cache type

        :param cache_type: enum representing the desired cache type
        :return: a new cache instance according to the requested cache type
        """
        if cache_type == Cache.Type.NONE:
            return DoNothingCache()
        elif cache_type == Cache.Type.REDIS:
            return RedisCache()
        elif cache_type == Cache.Type.IN_MEMORY:
            return InMemoryCache()
        else:
            raise AttributeError(f"Unsupported cache type {cache_type}")


class DoNothingCache(Cache):
    """!Empty implementation of the Cache interface. Does absolutely nothing."""

    def __init__(self) -> None:
        super().__init__()

    def status(self) -> bool:
        """!Empty implementation - returns True"""
        return True

    def get(self, keys: List[str]) -> List[Any]:
        """!Empty implementation - returns an empty list"""
        return []

    def set(self, cache_items: List[CacheItem]) -> None:
        """!Empty implementation - does nothing"""
        pass


class InMemoryCache(Cache):
    """!An in-memory implementation of the Cache interface.
    WARNING: This cache is not intended for production!!!
    It can be useful in testing and in cases where a quick out-of-the-box implementation is required.
    This cache has a size limit of 1 million items. When the limit is reached, the oldest item is deleted to make room
    for a new item.

    Attributes:
        max_size_bytes: a maximum cache size to prevent out-of-memory issue (500 MB by default).
    """

    def __init__(self, max_size_bytes: int = 500000000) -> None:  # 500 MB
        super().__init__()
        self.cache = collections.OrderedDict()
        self.max_size_bytes = max_size_bytes

    def status(self) -> bool:
        return True

    def get(self, keys: List[str]) -> List[Any]:
        retrieved_items = []
        for key in keys:
            if key in self.cache:
                retrieved_items.append(self.cache[key])
            else:
                retrieved_items.append(None)
        return retrieved_items

    def set(self, cache_items: List[CacheItem]) -> None:
        for cache_item in cache_items:
            while sys.getsizeof(self.cache) + sys.getsizeof(cache_item) > self.max_size_bytes:
                # Cache is full, remove the oldest item
                self.cache.popitem(last=False)
            self.cache[cache_item.key] = cache_item.value



from typing import Any

from dsframework.base.trainer.pl_wrapper import *

##
# @file
# @brief generatedClass class, define neural network layers.
class generatedClass(ZIDSNetwork):
    """! Define your networks here, or load from pretrained model.

    for example:
        @code
        model = nn.Sequential(
            nn.Linear(28 * 28, 64),
            nn.ReLU(),
            nn.Linear(64, 3),
            nn.Linear(3, 64),
            nn.ReLU(),
            nn.Linear(64, 28 * 28)
        )
        @endcode

    Another example:
        @code
        nnetwork_wine = nn.Sequential(
        nn.Linear(11, 1),
        nn.ReLU(),
        nn.Linear(1, 1))
        @endcode

    Another example:
        @code
        self.bert = BertModel.from_pretrained("bert-base-cased")
        @endcode
    """

    model = None

    def __init__(self):
        super().__init__(self.model)

    @staticmethod
    def load_define_model():
        """Load the model and tokenizer here, called from pytorch lightning model constructor.

        Example:
            model_name = 'distilbert-base-cased'
            model = AutoModelForTokenClassification.from_pretrained(model_name, num_labels=6)
            tokenizer = AutoTokenizer.from_pretrained(model_name)

            new_word = '<NEW_LINE>'
            tokenizer.add_tokens(new_word)
            model.resize_token_embeddings(len(tokenizer))
            return model, tokenizer

        Another example:
            model = nn.Sequential(
                nn.Linear(28 * 28, 64),
                nn.ReLU(),
                nn.Linear(64, 3),
                nn.Linear(3, 64),
                nn.ReLU(),
                nn.Linear(64, 28 * 28)
            )

        """
        model = None
        tokenizer = None

        if model is None:
            raise Exception("load_model() not defined. please define in network_module.py")

        return model, tokenizer



import os

from dsframework.base.trainer.train import ZIDSTrainClass
from dsframework.base.trainer.pl_wrapper import *
from trainer.pl_wrapper.save_onnx import ToOnnx

##
# @file
# @brief generatedClass class, executes training process.


class generatedClass(ZIDSTrainClass):
    """ Trainer class, responsible to execute training process and save the model."""

    def __init__(self, trainer_config=None, model_config=None, train_dataset=None, validation_dataset=None,
                 model=None, data_module=None, save_path=''):
        """ Train class initializer."""
        super().__init__(trainer_config, model_config, train_dataset, validation_dataset, model, data_module,
                         save_path)

    def execute(self):
        """ Execute the train process and saves the model.

        Needs to be implemented.

        Example (pytorch lightning):
            self.trainer = pl.Trainer(**self.trainer_config)
            self.trainer.fit(self.model, self.data_module)
        """

        if self.model_config['save_last_model']:
            self.save_model(os.path.join(self.save_path, self.model_config['model_name']))

        if self.model_config.get('save_checkpoint_as_onnx', False):
            checkpoint_path = self.trainer_config['callbacks'][0].best_model_path
            print(f'best model path: {checkpoint_path}')

            to_onnx = ToOnnx(model_config=self.model_config)
            to_onnx.save_checkpoint_as_onnx(checkpoint_path, self.model)

        if self.trainer is None:
            raise Exception('Implement train process in train.py, execute() method.')

    def save_model(self, model_path):
        """ Saves the model, runs at the end of the training process.

        Needs to be implemented.

        Notes:
            Saves the last model, which doesn't mean it is the best model. If using pytorch lightning, use
            ModelCheckpoint callback to save the best model, callback defined in config.py.

        Example (pytorch lightning):
            if self.trainer is not None:
                self.trainer.save_checkpoint(model_path)
                print(message)
        """

        message = 'Saves the last model, which doesn\'t mean it is the best model. If using pytorch lightning, ' \
                  'use ModelCheckpoint callback to save the best model, callback defined in config.py.'

        raise 'train.py save_model() not implemented. ' + message

    def get_trainer(self):
        """Returns trainer instance."""
        return self.trainer



import os

from dsframework.base.trainer.pl_wrapper import *


class ToOnnx:
    def __init__(self, model_config=None):

        self.model_config = model_config
        self.input_length = 0  # size of the input length.

    def save_checkpoint_as_onnx(self, checkpoint_path, pl_model_class):
        success = True
        try:
            model = self.load_model(checkpoint_path, pl_model_class)

            if self.input_length == 0:
                self.input_length = model._modules['0'].in_features  # set input length here

            input_ids, att_masks = self.generate_random_int_ids(self.input_length)

            save_to = os.path.splitext(checkpoint_path)[0] + '.onnx'
            success = self.save_to_onnx(model, save_to, input_ids, att_masks)

        except Exception as ex:
            print(f'Exception in save_checkpoint_as_onnx: {ex}')

        return success

    def load_model(self, checkpoint_path, pl_model_class):

        checkpoint_model = pl_model_class.load_from_checkpoint(
            checkpoint_path, strict=False,
            model_config=self.model_config,
            metrics_config=None,
            trainer_config=None)

        model = checkpoint_model.model.to(self.model_config.get('device', 'cpu'))
        model.eval()

        return model

    def generate_random_int_ids(self, input_length, range_from=0, range_to=3000):

        input_ids = torch.randint(range_from, range_to, (1, input_length), requires_grad=False)  # Simulates the input_ids
        att_masks = torch.randint(0, 2, (1, input_length), requires_grad=False)

        return input_ids, att_masks

    def save_to_onnx(self, model, save_to_path, input_ids, attention_mask):

        success = True

        try:
            torch.onnx.export(
                model,  # model being run
                (input_ids, attention_mask),  # model input (or a tuple for multiple inputs)
                save_to_path,  # where to save the model (can be a file or file-like object)
                export_params=True,  # store the trained parameter weights inside the model file
                opset_version=12,  # the ONNX version to export the model to -  recommended - the latest version
                do_constant_folding=True,  # whether to execute constant folding for optimization
                input_names=['input_ids', 'attention_mask'],  # the model's input names
                output_names=['output'],  # ,  # the model's output names
                dynamic_axes={'input_ids': {0: 'batch_size', 1: 'seq_len'},  # variable length axes
                              'attention_mask': {0: 'batch_size', 1: 'seq_len'},
                              'output': {0: 'batch_size'}})
        except Exception as ex:
            print(f'exception saving onnx: {ex}')
            success = False

        return success



class ZIDSTrainerComparison:

    comparison_results = {}

    test_results = None
    eval_results = None
    metrics = None  # ['precision', 'accuracy', 'recall', 'f1', 'confusion_matrix']

    def __init__(self, test_results, eval_results, metrics_config):

        self.test_results = test_results
        self.eval_results = eval_results
        self.metrics = metrics_config

    def compare(self):
        if self.test_results is None or self.eval_results is None:
            return "No test results found."

        self.comparison_results = {
            metric: {'trained': self.test_results[metric], 'prd': self.eval_results[metric]}
            for metric in self.metrics
            if metric in self.test_results and metric in self.eval_results and self.metrics[metric]}

        return self.comparison_results
    


import atexit
import os
import shutil
import sys
from typing import List
from glob import glob
from setuptools import setup
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.install import install

# Extract build type
build_type = ''
if sys.argv and len(sys.argv) > 2:
	build_type = sys.argv[2]
	index = sys.argv.index(sys.argv[2])
	sys.argv.pop(index)

# If a build folder exists, delete it
build_dir = './build/'
if os.path.isdir(build_dir):
	shutil.rmtree(build_dir)

# Extract required packages
main_project_path = os.getcwd()
with open(main_project_path + '/requirements_docker.txt') as f:
	required = f.read().splitlines()
# print('required', required)


def package_files(directory, exclude_dirs: List = None):
	paths = []
	for (path, directories, filenames) in os.walk(directory):
		if exclude_dirs and isinstance(exclude_dirs, list):
			for item in exclude_dirs:
				if item in directories:
					directories.remove(item)
		for filename in filenames:
			paths.append(os.path.join('../' + path, filename))
	return paths


def read(fname):
	return open(os.path.join(os.path.dirname(__file__), fname)).read()


project_name = '{name-your-service}'


class PostInstallCommand(install):
	"""Post-installation for installation mode."""
	def run(self):
		def _post_install():
			ext = '.whl'
			target_ext = '.zip'
			path = main_project_path + '/dist/'
			target_file_name = path + project_name + target_ext
			if os.path.exists(target_file_name):
				os.remove(target_file_name)
			fileName = os.path.basename(glob(path + "*" + ext)[0])
			target_file_name = path + project_name + target_ext
			os.rename(path + fileName, target_file_name)

		atexit.register(_post_install)
		install.run(self)


class PostInstallCommand2(bdist_egg):
	"""Post-installation for installation mode."""
	def run(self):
		def _post_install():
			ext = '.egg'
			target_ext = '.egg'
			path = main_project_path + '/dist/'
			target_file_name = path + project_name + target_ext
			if os.path.exists(target_file_name):
				os.remove(target_file_name)
			fileName = os.path.basename(glob(path + "*" + ext)[0])
			target_file_name = path + project_name + target_ext
			os.rename(path + fileName, target_file_name)

		atexit.register(_post_install)
		bdist_egg.run(self)

# Left as an example for creating more than one build type
# pipes = package_files('pipes', ['.dvc', 'py_signature_parser', '__pycache__'])


pipeline = package_files('pipeline', ['__pycache__'])
batch_files = package_files('batch_files', ['__pycache__'])
tester = package_files('tester', ['__pycache__'])
trainer = package_files('trainer', ['__pycache__', 'datasets', 'lightning_logs', 'training_output'])
config = package_files('config')

packages = ['pipeline']
data = config + pipeline + batch_files + tester
project_name += '_' + build_type

# Additional workflows go here. Add specific packages / data to your workflow type (Do not remove this comment)
if build_type == 'trainer':
	packages = ['trainer']
	data = config + pipeline + batch_files + tester + trainer

setup(
	name=project_name,
	packages=packages,
	package_data={
		'': [
				'../__init__.py',
		] + data
	},

	include_package_data=True,
	install_requires=required,
	cmdclass={
		'install': PostInstallCommand,
		'bdist_egg': PostInstallCommand2
	}
)
# commands
# for test -  python setup.py pytest
# for build wheel -  python setup.py bdist_wheel
# for source dist -  python setup.py sdist
# for build -  python setup.py build
# for install -  python setup.py install
# for uninstall - python -m pip uninstall trex-batch
# for install - python -m pip install dist/trex-batch-0.1.0-py3-none-any.whl

# deploy to PyPI
# delete dist and build folders
# python setup.py bdist_wheel
# python setup.py sdist
# python setup.py build
# twine upload dist/*
'''
	use
	1. python setup.py install
	2. dsf-cli g model new_model_name
	3. twine check dist/*
	4. twine upload --repository-url https://pypi.org/legacy/ dist/*
	4. twine upload dist/*
	
	pip install dsframework --index-url https://pypi.org/simple

	how to use

	pip install dsframework
	
	dsf-cli generate project my-new-model
'''



"""! @brief Skeleton for implementing a specific stage to be used in a DataProc Workflow template."""
import argparse
import sys
import logging
from json import loads as json_loads
from dsframework.base.batch.pipeline_stage_base import ZIDS_BatchPipelineStage

logger = logging.getLogger(__name__)


##
# @file
# @brief generatedStageName class, inherits ZIDS_BatchPipelineStage base class.
class generatedStageName(ZIDS_BatchPipelineStage):
    """! The stage class that will be later converted to an executable job in a specific workflow.
    See class ExampleStage for an example implementation of the load, run and finish methods.
    A method for initiating a SparkSession is available in the base class ZIDS_BatchPipelineStage
    For logging please use the module-level logger - 'logger'
    """

    def __init__(self, stage_config):
        """! The Stage class (generatedStageName) initializer.
        Base class will load basic configuration parameters, additional fields should be added here

            Args:
                stage_config : Configuration dictionary, loaded from configuration file.
        """
        self.wf_name = "generatedWFName"
        super().__init__(stage_config, logger)

    def load(self):
        """! The \'load\' phase is for loading input dataframes and running input data tests if needed
        """
        raise NotImplementedError

    def run(self):
        """! The \'run\' phase is for the transform logic
        """
        raise NotImplementedError

    def finish(self):
        """! The \'finish\' phase is for saving output dataframes and running output data tests if needed
        """
        raise NotImplementedError


if __name__ == "__main__":
    """! Executes the stage by instantiating it and calling the __call__ function of the base class.
        
        Args:
            System argument 1 - Configuration file
            System argument 2 - Start date, received from Airflow
            System argument 3 - End date, received from Airflow
            System argument 4 - Placeholder for future parameters
    """

    # Handle input arguments
    parser = argparse.ArgumentParser(description='Configure and run generatedStageName class')
    parser.add_argument('config', type=str,
                        help='Configuration file')
    parser.add_argument('start_date', type=str, default='',
                        help='Start date, received from Airflow')
    parser.add_argument('end_date', type=str, default='',
                        help='End date, received from Airflow')
    parser.add_argument('params', nargs=argparse.REMAINDER,
                        help='Placeholder for future parameters')
    args = parser.parse_args()

    # Configure stage class and run
    stage = generatedStageName(json_loads(args.config))
    try:
        stage.update_stage_params(args.start_date, args.end_date, args.params)
        stage()
    except Exception as e:
        raise Exception(f" Stage failed with error: {e}")






import requests
import json
from http import HTTPStatus
from pipeline.schema.inputs import generatedProjectNameInputs
from token_generator import tokenGenerator
##
# @file
# @brief test_server_post is used in order to check
#        the service liveness for example or post a request and get the data (response) back.
#        (only for checking use)

# select local or staging
env = 'local'

if env == 'local':
    base_url = 'http://localhost:8080/'
    jwtToken = 'dummy_jwt_token'
elif env == 'staging':
    base_url = 'https://<project staging environment url>/'
    t = tokenGenerator()
    jwtToken = t.generateToken('STG_Conf')
else:
    raise AttributeError(f'Invalid environment {env}')

global_headers = {
    'x-token': 'fake-super-secret-token',
    'Authorization': 'Bearer ' + jwtToken
}


def post(end_point, data):
    """! Post method, send post request to a specific URL and return the response back in the data.

     Args:
        data : json file data that will use for the request post.
        end_point : end point of the URL. (base_url + endPoint will be the complete url)
     Returns:
        data : Response content
     """
    headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
    url = base_url + end_point
    response = requests.post(url, json=data, headers=headers)
    print(f"Response elapsed time: {response.elapsed.total_seconds()}")
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        data = HTTPStatus.INTERNAL_SERVER_ERROR.description
    else:
        try:
            data = json.loads(response.content)
        except Exception as ex:
            data = response.content
    return data


def get(end_point, data):
    """! Get method, send get request to a specific URL and return the response back.

     Args:
        data : data params that will use for the get request
        end_point : end point of the URL. (base_url + endPoint will be the complete url)
     Returns:
        data : Response content
     """
    headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
    url = base_url + end_point
    response = requests.get(url, params=data, headers=headers)
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        data = HTTPStatus.INTERNAL_SERVER_ERROR.description
    else:
        try:
            data = json.loads(response.content)
        except Exception as ex:
            data = response.content
    return data


if __name__ == '__main__':
    # test - parse
    data = {}
    d = generatedProjectNameInputs.parse_obj(data)

    results = get('livenessprobe', {})
    print('results', results)

    results = post('predict', d.dict())
    print('results', results)



from dsframework.base.trainer.pl_wrapper import *
from abc import ABC, abstractmethod


class ModuleBase(ABC, nn.Module):
    """! Base class for a module inherits from torch nn.module."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def forward(self, x):
        pass
